#include <cstdlib>
#include <stdio.h>
#include <string.h>
#include "GPIOPin.h"
#include "GPIOAccess.h"

using namespace std;

#define numPins 27

void usage(char * nm) {
    printf("%s\nUsage\n", nm);
    printf("\t%s <op> <pin> <val>\n",nm);
    printf("Where:\n");
    printf("\t<op> is one of:\n");
    printf("\t\tlist - to list info\n");
    printf("\t\tset - to set pin value\n");
    printf("\t\tget - to get and return pin value\n");
    printf("\t\tsetd - to set pin direction\n");
    printf("\t\tgetd - to get and return pin direction\n");
    printf("\t\thelp - to display usage\n");
    printf("\t<pin> is one of\n\t\t0, 1, 6, 7, 8, 12, 13, 14, 15, 16, 17, 18, 19, 23, 26, all\n");
    printf("\t\tA <pin> of all can only be used for an <op> of list, set or setd\n");
    printf("\t<val> is only required for set and setd:\n");
    printf("\t\tfor set, <val> is 0 or 1\n");
    printf("\t\tfor setd, <val> is in or out\n");
}

#define oplist 0
#define opset 1
#define opget 2
#define opsetd 3
#define opgetd 4
int operation;
int pinnum;
int value;
GPIO_Direction direction;

bool processArgs(int argc, char** argv) {
    operation = -1;
    pinnum = -1;
    value = -1;
    if (argc > 1) {
        if (strcmp(argv[1], "list") == 0) {
            operation = oplist;
        } else if (strcmp(argv[1], "set") == 0) {
            operation = opset;
        } else if (strcmp(argv[1], "get") == 0) {
            operation = opget;
        } else if (strcmp(argv[1], "setd") == 0) {
            operation = opsetd;
        } else if (strcmp(argv[1], "getd") == 0) {
            operation = opgetd;
        } else if (strcmp(argv[1], "help") == 0) {
            return false;
        } else {
            printf("**ERROR** Invalid <op> : %s\n", argv[1]);
            return false;
        }
        
        if (argc > 2) {
            if (strcmp(argv[2], "all") == 0) {
                if ((operation == opget) || (operation == opgetd)) {
                    printf("**ERROR** Invalid <pin> for operation: %s\n", argv[2]);
                    return false;
                }
                pinnum = -1;
            } else if (strcmp(argv[2], "0") == 0) {
                pinnum = 0;
            } else {
                pinnum = strtol(argv[2], NULL, 10);
                if (pinnum != 0) {
                    if (!GPIOAccess::isPinUsable(pinnum)) {
                        printf("**ERROR** Invalid <pin> : %s\n", argv[2]);
                        return false;
                    }
                } else {
                    printf("**ERROR** Invalid <pin> : %s\n", argv[2]);
                    return false;
                }
            }
            
            if (operation == opset) {
                if (argc > 3) {
                    if (strcmp(argv[3], "0") == 0) {
                        value = 0;
                    } else if (strcmp(argv[3], "1") == 0 ) {
                        value = 1;
                    } else{
                        printf("**ERROR** Invalid <val> for set: %s\n", argv[3]);
                        return false;
                    }
                } else {
                    printf("**ERROR** No <val> specified for set\n");
                    return false;
                }
            } else if (operation == opsetd) {
                if (argc > 3) {
                    if (strcmp(argv[3], "in") == 0) {
                        direction = GPIO_INPUT;
                    } else if (strcmp(argv[3], "out") == 0 ) {
                        direction = GPIO_OUTPUT;
                    } else{
                        printf("**ERROR** Invalid <val> for setd: %s\n", argv[3]);
                        return false;
                    }
                } else {
                    printf("**ERROR** No <val> specified for setd\n");
                    return false;
                }
            }
        } else {
            printf("**ERROR** No <pin> specified\n");
            return false;
        }
    } else {
        printf("**ERROR** No <op> specified\n");
        return false;
    }
    
    return true;
}
void setAll(GPIO_Direction dir) {
    int i;
    for (i = 0; i < numPins; i++) {
        if (!GPIOAccess::isPinUsable(i)) {
            continue;
        }
        GPIOPin *pin = new GPIOPin(i);
        pin->setDirection(dir);
        delete pin;
    }
}

void reportOn(int pinnum) {
    printf("Pin:%d, ", pinnum);
    GPIOPin *pin = new GPIOPin(pinnum);
    GPIO_Direction dir;
    pin->getDirection(dir);
    printf("Direction:");
    if (dir == GPIO_OUTPUT) {
        printf("OUTPUT");
    } else if (dir == GPIO_INPUT) {
        int val;
        pin->get(val);
        printf("INPUT, Value:%d", val);
    }
    printf("\n");
    delete pin;
}

void setPin(int pinnum, int val) {
    GPIOPin *pin = new GPIOPin(pinnum);
    GPIO_Result r = pin->set(val);
    delete pin;
    if (r != GPIO_OK) {
        printf("**ERROR setting pin:%d to value:%d, err=%d\n", pinnum, val, r);
    }
}

void setPinDir(int pinnum, GPIO_Direction dir) {
    GPIOPin *pin = new GPIOPin(pinnum);
    GPIO_Result r = pin->setDirection(dir);
    delete pin;
    if (r != GPIO_OK) {
        printf("**ERROR setting pin direction:%d to direction:%d, err=%d\n", pinnum, dir, r);
    }
}

GPIO_Result getPin(int pinnum, int &val) {
    GPIOPin *pin = new GPIOPin(pinnum);
    GPIO_Result r = pin->get(val);
    delete pin;
    if (r != GPIO_OK) {
        printf("**ERROR getting pin:%d value, err=%d\n", pinnum, r);
    }
    return r;
}

GPIO_Result getPinDir(int pinnum, GPIO_Direction &dir) {
    GPIOPin *pin = new GPIOPin(pinnum);
    GPIO_Result r = pin->getDirection(dir);
    delete pin;
    if (r != GPIO_OK) {
        printf("**ERROR getting pin:%d direction, err=%d\n", pinnum, r);
    }
    return r;
}

int main(int argc, char** argv) {
    if (!processArgs(argc, argv)) {
        usage(argv[0]);
        return -1;
    }
    
    if (operation == oplist) {
        if (pinnum != -1) {
            reportOn(pinnum);
        } else {
            int i;
            for (i = 0; i < numPins; i++) {
                if (GPIOAccess::isPinUsable(i)) {
                    reportOn(i);
                }
            }
        }
        return 0;
    } else if (operation == opset) {
        if (pinnum != -1) {
            setPin(pinnum, value);
        } else {
            int i;
            for (i = 0; i < numPins; i++) {
                if (GPIOAccess::isPinUsable(i)) {
                    setPin(i, value);
                }
            }
        }
        return 0;
    } else if (operation == opget) {
        int val;
        GPIO_Result r = getPin(pinnum, val);
        if (r == GPIO_OK) {
            return val;
        } else {
            return r;
        }
    } else if (operation == opsetd) {
        if (pinnum != -1) {
            setPinDir(pinnum, direction);
        } else {
            int i;
            for (i = 0; i < numPins; i++) {
                if (GPIOAccess::isPinUsable(i)) {
                    setPinDir(i, direction);
                }
            }
        }
        return 0;
    } else if (operation == opgetd) {
        GPIO_Direction dir;
        GPIO_Result r = getPinDir(pinnum, dir);
        if (r == GPIO_OK) {
            return dir;
        } else {
            return r;
        }
    }
    
    return -1;
}

